!reheating functions for the large field model with massive curvaton

  module mchi9reheat

    use infprec, only : kp, tolkp, transfert
    use inftools, only : zbrent
    use cosmopar, only : powerAmpScalar
    use mchisr, only : mchi_potential, mchi_deriv_second_potential, mchi_phiend0, &
        mchi_phi_trajectory, mchi_deriv_potential, mchi_InverseVp_primitive, &
        mchi_phi_trajectory, mchi_deriv_third_potential, mchi_sigma_trajectory, &
        mchi_Delta_faciq
    use curvaton, only : find_MC_bfoldstar, MC9_lnRrad, MC9_Pzeta, MC_Delta, MC_inftraj_sigma, &
                         MC_r, MC_ns, MC_alphaS, MC9_fNL_phistarfree, MC9_Mmin_reh, &
                         MC9_Pzeta_relativecontribution_sigma

    implicit none

    private

    public mchi9_ns, mchi9_r, mchi9_fnl, mchi9_isnormalizable, mchi9_M, mchi9_bfoldstar, &
            mchi9_Pzeta_sigma, mchi9_alphaS, mchi9_sigma_trajectory, mchi9_ns_faciq, &
           mchi9_r_faciq, mchi9_fNL_faciq, mchi9_alphaS_faciq, mchi9_phi_trajectory

  contains


!!!!!!!!!!!!!!!!!!!!!!
!! Number of efolds !!
!!!!!!!!!!!!!!!!!!!!!!

!Returns the number of bfolds given potential parameters
  function mchi9_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mchi9_bfoldstar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp), parameter :: tohind=tolkp
    real(kp) :: bfoldstarmini,bfoldstarmaxi,bfoldstar
    type(transfert) :: mchiData

    mchiData%real1 = Vparam%real1 ! M

    mchiData%real3 = msigma
    mchiData%real4 = Gammaphi
    mchiData%real5 = Gammasigma
    mchiData%real6 = sigmaend

    bfoldstarmini = -1000._kp
    bfoldstarmaxi = 0._kp

    bfoldstar = zbrent(find_mchi9_bfoldstar,bfoldstarmini,bfoldstarmaxi,tohind,mchiData)

    mchi9_bfoldstar = bfoldstar

  end function mchi9_bfoldstar

! Returns the equation one must solve to obtain bfoldstar from model parameters
  function find_mchi9_bfoldstar(bfoldstar,mchiData)
    implicit none
    real(kp) :: find_mchi9_bfoldstar
    real(kp), intent(in) :: bfoldstar
    type(transfert), optional, intent(inout) :: mchiData
    type(transfert) :: Vparam

    real(kp) :: msigma, Gammaphi, Gammasigma, sigmaend, lnRrad, Pstar, phistar

    Vparam%real1 = mchiData%real1! M

    msigma = mchiData%real3
    Gammaphi = mchiData%real4
    Gammasigma = mchiData%real5
    sigmaend = mchiData%real6

    lnRrad = MC9_lnRrad(mchi_potential, mchi_deriv_second_potential, mchi_phiend0, Vparam, &
             msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mchi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    Pstar = MC9_Pzeta(mchi_potential, mchi_deriv_potential, mchi_deriv_second_potential, &
            mchi_phiend0, mchi_InverseVp_primitive, Vparam, msigma, Gammaphi, Gammasigma, &
            sigmaend, phistar)

    find_mchi9_bfoldstar = find_MC_bfoldstar(mchi_potential, mchi_deriv_potential, &
         mchi_deriv_second_potential, mchi_phiend0, mchi_InverseVp_primitive, &
         mchi_phi_trajectory, Vparam, msigma, Gammaphi, Gammasigma, sigmaend, lnRrad, &
         Pstar, bfoldstar)

  end function find_mchi9_bfoldstar

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Power Spectrum Amplitude  !!
!!      Normalization         !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! Returns the power spectrum amplitude once the number of bfolds_star has been solved
  function mchi9_Pstar_bfoldssolved(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    real(kp) :: mchi9_Pstar_bfoldssolved
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: bfoldstar, phistar

    bfoldstar = mchi9_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mchi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    mchi9_Pstar_bfoldssolved = MC9_Pzeta(mchi_potential, mchi_deriv_potential, &
        mchi_deriv_second_potential,mchi_phiend0, mchi_InverseVp_primitive, Vparam, msigma, &
        Gammaphi, Gammasigma, sigmaend, phistar)

  end function mchi9_Pstar_bfoldssolved

! Returns the equation one must solve to obtain M from other model parameters by normalizing it to the scalar power spectrum amplitude
  function find_mchi9_M(M,mchiData)
    implicit none
    real(kp) :: find_mchi9_M
    real(kp), intent(in) :: M
    type(transfert), optional, intent(inout) :: mchiData
    real(kp) :: msigma, Gammaphi, Gammasigma, sigmaend, Pstar
    type(transfert) :: Vparam

    Vparam%real1 = M ! M

    msigma = mchiData%real3
    Gammaphi = mchiData%real4
    Gammasigma = mchiData%real5
    sigmaend = mchiData%real6

    Pstar = mchi9_Pstar_bfoldssolved(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)

    find_mchi9_M = Pstar - powerAmpScalar

  end function find_mchi9_M

!Returns the mass normalization of the inflaton potential (M) to the power spectrum amplitude
  function mchi9_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mchi9_M
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp), parameter :: tohind=tolkp
    real(kp) :: Mmini1, Mmini2, Mmini3, Mmini4, Mmini, Mmaxi, M
    type(transfert) :: mchiData

    mchiData%real3 = msigma
    mchiData%real4 = Gammaphi
    mchiData%real5 = Gammasigma
    mchiData%real6 = sigmaend

    Vparam%real1 = 1._kp ! M (dummy value, to get V/M^4 in Mmini)

    Mmini1 = (msigma**2*sigmaend**2/(3._kp* &
              mchi_potential(mchi_phiend0(Vparam),Vparam)))**(0.25_kp)*100._kp ! This is to have phi dominate the energy budget at the end of inflation, 100 is a safety number
    Mmini2 = (msigma**2/ &
              mchi_potential(mchi_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have sigma still light at the end of inflation
    Mmini3 = (2._kp*Gammaphi**2/ mchi_potential(mchi_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have Hend>Gammaphi
    Mmini4 = MC9_Mmin_reh(mchi_potential, mchi_deriv_second_potential, mchi_phiend0, &
            Vparam, msigma, Gammaphi, Gammasigma, sigmaend)! CMB pivot scale exits Hubble radius during the first inflationary phase

    Mmini = max(Mmini1,Mmini2,Mmini3,Mmini4)*10._kp !10 is a safety number
    Mmaxi = 1._kp

    if (find_mchi9_M(Mmini,mchiData)*find_mchi9_M(Mmaxi,mchiData) > 0._kp) then
         stop 'normalization to Power Spectrum Amplitude is Impossible'
    endif

    if (Mmini > Mmaxi) then
         stop 'Mmini>Mmaxi'
    endif

    M = zbrent(find_mchi9_M,Mmini,Mmaxi,tohind,mchiData)

    mchi9_M = M

  end function mchi9_M

!Test Function to determine if the power spectrum is normalizable
  function mchi9_isnormalizable(msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    logical :: mchi9_isnormalizable
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam
    real(kp) :: Mmini1, Mmini2, Mmini3, Mmini4, Mmini, Mmaxi
    type(transfert) :: mchiData

    Vparam%real1 = 1._kp ! M (dummy value, to get V/M^4 in Mmini)

    mchi9_isnormalizable = .false.
    Mmini1 = (msigma**2*sigmaend**2/(3._kp* &
              mchi_potential(mchi_phiend0(Vparam),Vparam)))**(0.25_kp)*100._kp ! This is to have phi dominate the energy budget at the end of inflation, 100 is a safety number
    Mmini2 = (msigma**2/ &
              mchi_potential(mchi_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have sigma still light at the end of inflation
    Mmini3 = (2._kp*Gammaphi**2/ mchi_potential(mchi_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have Hend>Gammaphi
    Mmini4 = MC9_Mmin_reh(mchi_potential, mchi_deriv_second_potential, mchi_phiend0, &
            Vparam, msigma, Gammaphi, Gammasigma, sigmaend)! CMB pivot scale exits Hubble radius during the first inflationary phase
    Mmini = max(Mmini1,Mmini2,Mmini3,Mmini4)*10._kp !10 is a safety number
    Mmaxi = 1._kp

    mchiData%real3 = msigma
    mchiData%real4 = Gammaphi
    mchiData%real5 = Gammasigma
    mchiData%real6 = sigmaend

    if (Mmini<Mmaxi) then
      if (find_mchi9_M(Mmini,mchiData)*find_mchi9_M(Mmaxi,mchiData) < 0._kp) then
        mchi9_isnormalizable = .true.
      endif
    endif

  end function mchi9_isnormalizable

  !Returns the correlation angle Delta between curvature and isocurvature power spectra
  function mchi9_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mchi9_Delta, mphi, bfoldstar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    type(transfert) :: Vparam_passed

    Vparam_passed%real1 = mchi9_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) ! Normalizes M

    bfoldstar = mchi9_bfoldstar(Vparam_passed, msigma, Gammaphi, Gammasigma, sigmaend)
    mchi9_Delta = MC_Delta(mchi_potential, mchi_deriv_potential, mchi_deriv_second_potential, &
                 mchi_phiend0, mchi_InverseVp_primitive, mchi_phi_trajectory, Vparam_passed, &
                 msigma, Gammaphi, Gammasigma, sigmaend, bfoldstar)

  end function mchi9_Delta

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Observable Predictions   !!
!!    Direct Calculation     !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  function mchi9_ns(msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mchi9_ns, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real1 = mchi9_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mchi9_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mchi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mchi_potential,mchi_deriv_second_potential,mchi_phiend0, &
                mchi_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mchi9_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mchi9_ns = MC_ns(mchi_potential,mchi_deriv_potential,mchi_deriv_second_potential, &
                mchi_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mchi9_ns

  function mchi9_alphaS(msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mchi9_alphaS, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real1 = mchi9_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mchi9_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mchi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mchi_potential,mchi_deriv_second_potential,mchi_phiend0, &
                mchi_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mchi9_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mchi9_alphaS = MC_alphaS(mchi_potential,mchi_deriv_potential,mchi_deriv_second_potential, &
                mchi_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mchi9_alphaS

  function mchi9_r(msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mchi9_r, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real1 = mchi9_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mchi9_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mchi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mchi_potential,mchi_deriv_second_potential,mchi_phiend0, &
                mchi_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mchi9_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mchi9_r = MC_r(mchi_potential,mchi_deriv_potential,mchi_deriv_second_potential,Vparam, &
               phistar,sigmastar,msigma,Delta)

  end function mchi9_r

  function mchi9_fNL(msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mchi9_fNL, bfoldstar, phistar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real1 = mchi9_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mchi9_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mchi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

    mchi9_fNL=MC9_fNL_phistarfree(mchi_potential,mchi_deriv_potential, &
               mchi_deriv_second_potential, mchi_phiend0, mchi_InverseVp_primitive, &
               Vparam, msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mchi9_fNL

! Returns the relative contribution of sigma to the total curvature power spectrum
  function mchi9_Pzeta_sigma(msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mchi9_Pzeta_sigma, bfoldstar, phistar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam


    Vparam%real1 = mchi9_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mchi9_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mchi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

    mchi9_Pzeta_sigma=MC9_Pzeta_relativecontribution_sigma(mchi_potential,mchi_deriv_potential, &
                      mchi_deriv_second_potential,mchi_phiend0,mchi_InverseVp_primitive,Vparam, &
                      msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mchi9_Pzeta_sigma

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Observable Predictions   !!
!! Step by Step Calculation  !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!returns phistar once bfoldstar is computed
  function mchi9_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    implicit none
    real(kp), intent(in) :: bfoldstar, msigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: mchi9_phi_trajectory

    mchi9_phi_trajectory = mchi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

  end function mchi9_phi_trajectory

!returns sigmastar once phistar is computed
  function mchi9_sigma_trajectory(phistar, Vparam, msigma, sigmaend)
    implicit none
    real(kp), intent(in) :: phistar, msigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: mchi9_sigma_trajectory

    mchi9_sigma_trajectory = mchi_sigma_trajectory(phistar, Vparam, msigma, sigmaend)

  end function mchi9_sigma_trajectory

!returns ns from already computed intermediary quantites (faciq)
  function mchi9_ns_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mchi9_ns_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mchi_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mchi9_ns_faciq = MC_ns(mchi_potential,mchi_deriv_potential,mchi_deriv_second_potential, &
                mchi_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mchi9_ns_faciq

!returns r from already computed intermediary quantites (faciq)
  function mchi9_r_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mchi9_r_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mchi_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mchi9_r_faciq = MC_r(mchi_potential,mchi_deriv_potential,mchi_deriv_second_potential, &
                Vparam,phistar,sigmastar,msigma,Delta)

  end function mchi9_r_faciq

!returns fNL from already computed intermediary quantites (faciq)
  function mchi9_fNL_faciq(msigma, Gammaphi, Gammasigma, sigmaend, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mchi9_fNL_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mchi_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mchi9_fNL_faciq = MC9_fNL_phistarfree(mchi_potential,mchi_deriv_potential, &
               mchi_deriv_second_potential, mchi_phiend0, mchi_InverseVp_primitive, &
               Vparam, msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mchi9_fNL_faciq

!returns alphaS from already computed intermediary quantites (faciq)
  function mchi9_alphaS_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mchi9_alphaS_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mchi_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mchi9_alphaS_faciq = MC_alphaS(mchi_potential,mchi_deriv_potential,mchi_deriv_second_potential, &
                mchi_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mchi9_alphaS_faciq



  end module mchi9reheat
