!reheating functions for the large field model with massive curvaton

  module mclfi6reheat

    use infprec, only : kp, tolkp, transfert
    use inftools, only : zbrent
    use cosmopar, only : powerAmpScalar
    use mclfisr, only : mclfi_potential, mclfi_deriv_second_potential, mclfi_phiend0, &
        mclfi_phi_trajectory, mclfi_deriv_potential, mclfi_InverseVp_primitive, &
        mclfi_phi_trajectory, mclfi_deriv_third_potential, mclfi_Delta_faciq, &
        mclfi_sigma_trajectory
    use curvaton, only : find_MC_bfoldstar, MC6_lnRrad, MC6_Pzeta, MC_Delta, MC_inftraj_sigma, &
                         MC_r, MC_ns, MC6_fNL_phistarfree, MC6_Mmin_reh, &
                        MC6_Pzeta_relativecontribution_sigma, MC_alphaS, MC6_sigmaend_max

    implicit none

    private

    public mclfi6_ns, mclfi6_r, mclfi6_fnl, mclfi6_isnormalizable, mclfi6_M, &
            mclfi6_bfoldstar, mclfi6_Pzeta_sigma, mclfi6_alphaS, mclfi6_sigma_trajectory, mclfi6_ns_faciq, &
           mclfi6_r_faciq, mclfi6_fNL_faciq, mclfi6_alphaS_faciq, mclfi6_phi_trajectory

  contains


!!!!!!!!!!!!!!!!!!!!!!
!! Number of efolds !!
!!!!!!!!!!!!!!!!!!!!!!

!Returns the number of bfolds given potential parameters
  function mclfi6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mclfi6_bfoldstar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp), parameter :: tolFind=tolkp
    real(kp) :: bfoldstarmini,bfoldstarmaxi,bfoldstar
    type(transfert) :: mclfiData

    mclfiData%real1 = Vparam%real1 ! M
    mclfiData%real2 = Vparam%real2 ! p
    mclfiData%real3 = msigma
    mclfiData%real4 = Gammaphi
    mclfiData%real5 = Gammasigma
    mclfiData%real6 = sigmaend

    bfoldstarmini = -1000._kp
    bfoldstarmaxi = 0._kp

    bfoldstar = zbrent(find_mclfi6_bfoldstar,bfoldstarmini,bfoldstarmaxi,tolFind,mclfiData)

    mclfi6_bfoldstar = bfoldstar

  end function mclfi6_bfoldstar

! Returns the equation one must solve to obtain bfoldstar from model parameters
  function find_mclfi6_bfoldstar(bfoldstar,mclfiData)
    implicit none
    real(kp) :: find_mclfi6_bfoldstar
    real(kp), intent(in) :: bfoldstar
    type(transfert), optional, intent(inout) :: mclfiData
    type(transfert) :: Vparam

    real(kp) :: msigma, Gammaphi, Gammasigma, sigmaend, lnRrad, Pstar, phistar

    Vparam%real1 = mclfiData%real1! M
    Vparam%real2 = mclfiData%real2 ! p

    msigma = mclfiData%real3
    Gammaphi = mclfiData%real4
    Gammasigma = mclfiData%real5
    sigmaend = mclfiData%real6

    lnRrad = MC6_lnRrad(mclfi_potential, mclfi_deriv_second_potential, mclfi_phiend0, Vparam, &
             msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mclfi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    Pstar = MC6_Pzeta(mclfi_potential, mclfi_deriv_potential, mclfi_deriv_second_potential, &
            mclfi_phiend0, mclfi_InverseVp_primitive, Vparam, msigma, Gammaphi, Gammasigma, &
            sigmaend, phistar)

    find_mclfi6_bfoldstar = find_MC_bfoldstar(mclfi_potential, mclfi_deriv_potential, &
         mclfi_deriv_second_potential, mclfi_phiend0, mclfi_InverseVp_primitive, &
         mclfi_phi_trajectory, Vparam, msigma, Gammaphi, Gammasigma, sigmaend, lnRrad, &
         Pstar, bfoldstar)

  end function find_mclfi6_bfoldstar

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Power Spectrum Amplitude  !!
!!      Normalization         !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! Returns the power spectrum amplitude once the number of bfolds_star has been solved
  function mclfi6_Pstar_bfoldssolved(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    real(kp) :: mclfi6_Pstar_bfoldssolved
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: bfoldstar, phistar

    bfoldstar = mclfi6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mclfi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    mclfi6_Pstar_bfoldssolved = MC6_Pzeta(mclfi_potential, mclfi_deriv_potential, &
        mclfi_deriv_second_potential,mclfi_phiend0, mclfi_InverseVp_primitive, Vparam, msigma, &
        Gammaphi, Gammasigma, sigmaend, phistar)

  end function mclfi6_Pstar_bfoldssolved

! Returns the equation one must solve to obtain M from other model parameters by normalizing it to the scalar power spectrum amplitude
  function find_mclfi6_M(M,mclfiData)
    implicit none
    real(kp) :: find_mclfi6_M
    real(kp), intent(in) :: M
    type(transfert), optional, intent(inout) :: mclfiData
    real(kp) :: msigma, Gammaphi, Gammasigma, sigmaend, Pstar
    type(transfert) :: Vparam

    Vparam%real1 = M ! M
    Vparam%real2 = mclfiData%real2 ! p

    msigma = mclfiData%real3
    Gammaphi = mclfiData%real4
    Gammasigma = mclfiData%real5
    sigmaend = mclfiData%real6

    Pstar = mclfi6_Pstar_bfoldssolved(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)

    find_mclfi6_M = Pstar - powerAmpScalar

  end function find_mclfi6_M

!Returns the mass normalization of the inflaton potential (M) to the power spectrum amplitude
  function mclfi6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mclfi6_M
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp), parameter :: tolFind=tolkp
    real(kp) :: Mmini1, Mmini2, Mmini3, Mmini, Mmaxi, M
    type(transfert) :: mclfiData

    mclfiData%real2 = Vparam%real2 ! p
    mclfiData%real3 = msigma
    mclfiData%real4 = Gammaphi
    mclfiData%real5 = Gammasigma
    mclfiData%real6 = sigmaend

    Vparam%real1 = 1._kp ! M (dummy value, to get V/M^4 in Mmini)

    Mmini1 = (msigma**2*sigmaend**2/(3._kp* &
              mclfi_potential(mclfi_phiend0(Vparam),Vparam)))**(0.25_kp)*100._kp ! This is to have phi dominate the energy budget at the end of inflation, 100 is a safety number
    Mmini2 = (msigma**2/ &
              mclfi_potential(mclfi_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have sigma still light at the end of inflation
    Mmini3 = MC6_Mmin_reh(mclfi_potential, mclfi_deriv_second_potential, mclfi_phiend0, &
            Vparam, msigma, Gammaphi, Gammasigma, sigmaend)! CMB pivot scale exits Hubble radius during the first inflationary phase

    Mmini = max(Mmini1,Mmini2,Mmini3)*10._kp !10 is a safety number
    Mmaxi = 1._kp

    if (find_mclfi6_M(Mmini,mclfiData)*find_mclfi6_M(Mmaxi,mclfiData) > 0._kp) then
         stop 'normalization to Power Spectrum Amplitude is Impossible'
    endif

    if (Mmini > Mmaxi) then
         stop 'Mmini>Mmaxi'
    endif

    M = zbrent(find_mclfi6_M,Mmini,Mmaxi,tolFind,mclfiData)

    mclfi6_M = M

  end function mclfi6_M

!Test Function to determine if the power spectrum is normalizable
  function mclfi6_isnormalizable(p, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    logical :: mclfi6_isnormalizable
    real(kp), intent(in) :: p, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam
    real(kp) :: Mmini1, Mmini2, Mmini3, Mmini, Mmaxi, M
    type(transfert) :: mclfiData

    Vparam%real1 = 1._kp ! M (dummy value, to get V/M^4 in Mmini)
    Vparam%real2 = p

    mclfi6_isnormalizable = .false.
    Mmini1 = (msigma**2*sigmaend**2/(3._kp* &
              mclfi_potential(mclfi_phiend0(Vparam),Vparam)))**(0.25_kp)*100._kp ! This is to have phi dominate the energy budget at the end of inflation, 100 is a safety number
    Mmini2 = (msigma**2/ &
              mclfi_potential(mclfi_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have sigma still light at the end of inflation
    Mmini3 = MC6_Mmin_reh(mclfi_potential, mclfi_deriv_second_potential, mclfi_phiend0, Vparam, &
             msigma, Gammaphi, Gammasigma, sigmaend)! CMB pivot scale exits Hubble radius during the first inflationary phase
    Mmini = max(Mmini1,Mmini2,Mmini3)*10._kp !10 is a safety number
    Mmaxi = 1._kp
    mclfiData%real2 = Vparam%real2 ! p
    mclfiData%real3 = msigma
    mclfiData%real4 = Gammaphi
    mclfiData%real5 = Gammasigma
    mclfiData%real6 = sigmaend

    if (Mmini<Mmaxi) then
      if (find_mclfi6_M(Mmini,mclfiData)*find_mclfi6_M(Mmaxi,mclfiData) < 0._kp) then
        M = mclfi6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
        if (sigmaend < MC6_sigmaend_max(mclfi_potential,mclfi_phiend0,Vparam,msigma,Gammaphi,Gammasigma)) then ! check that sigmaend does not domimate the energy content at the end of inflation
          mclfi6_isnormalizable = .true.
        endif
      endif
    endif

  end function mclfi6_isnormalizable

  !Returns the correlation angle Delta between curvature and isocurvature power spectra
  function mclfi6_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mclfi6_Delta, mphi, bfoldstar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    type(transfert) :: Vparam_passed

    Vparam_passed%real1 = mclfi6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) ! Normalizes M
    Vparam_passed%real2 = Vparam%real2 ! Passes p
    bfoldstar = mclfi6_bfoldstar(Vparam_passed, msigma, Gammaphi, Gammasigma, sigmaend)
    mclfi6_Delta = MC_Delta(mclfi_potential, mclfi_deriv_potential, mclfi_deriv_second_potential, &
                 mclfi_phiend0, mclfi_InverseVp_primitive, mclfi_phi_trajectory, Vparam_passed, &
                 msigma, Gammaphi, Gammasigma, sigmaend, bfoldstar)

  end function mclfi6_Delta

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Observable Predictions   !!
!!    Direct Calculation     !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  function mclfi6_ns(p, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mclfi6_ns, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: p, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = p !passes M
    Vparam%real1 = mclfi6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mclfi6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mclfi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mclfi_potential,mclfi_deriv_second_potential,mclfi_phiend0, &
                mclfi_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mclfi6_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mclfi6_ns = MC_ns(mclfi_potential,mclfi_deriv_potential,mclfi_deriv_second_potential, &
                mclfi_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mclfi6_ns

  function mclfi6_alphaS(p, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mclfi6_alphaS, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: p, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = p !passes M
    Vparam%real1 = mclfi6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mclfi6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mclfi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mclfi_potential,mclfi_deriv_second_potential,mclfi_phiend0, &
                mclfi_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mclfi6_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mclfi6_alphaS = MC_alphaS(mclfi_potential,mclfi_deriv_potential,mclfi_deriv_second_potential, &
                mclfi_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mclfi6_alphaS

  function mclfi6_r(p, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mclfi6_r, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: p, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = p !passes p
    Vparam%real1 = mclfi6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mclfi6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mclfi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mclfi_potential,mclfi_deriv_second_potential,mclfi_phiend0, &
                mclfi_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mclfi6_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mclfi6_r = MC_r(mclfi_potential,mclfi_deriv_potential,mclfi_deriv_second_potential,Vparam, &
               phistar,sigmastar,msigma,Delta)

  end function mclfi6_r

  function mclfi6_fNL(p, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mclfi6_fNL, bfoldstar, phistar
    real(kp), intent(in) :: p, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = p !passes p
    Vparam%real1 = mclfi6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mclfi6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mclfi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

    mclfi6_fNL=MC6_fNL_phistarfree(mclfi_potential,mclfi_deriv_potential, &
               mclfi_deriv_second_potential, mclfi_phiend0, mclfi_InverseVp_primitive, &
               Vparam, msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mclfi6_fNL

! Returns the relative contribution of sigma to the total curvature power spectrum
  function mclfi6_Pzeta_sigma(p, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mclfi6_Pzeta_sigma, bfoldstar, phistar
    real(kp), intent(in) :: p, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = p !passes p
    Vparam%real1 = mclfi6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mclfi6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mclfi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

    mclfi6_Pzeta_sigma=MC6_Pzeta_relativecontribution_sigma(mclfi_potential,mclfi_deriv_potential, &
                      mclfi_deriv_second_potential,mclfi_phiend0,mclfi_InverseVp_primitive,Vparam, &
                      msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mclfi6_Pzeta_sigma

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Observable Predictions   !!
!! Step by Step Calculation  !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!returns phistar once bfoldstar is computed
  function mclfi6_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    implicit none
    real(kp), intent(in) :: bfoldstar, msigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: mclfi6_phi_trajectory

    mclfi6_phi_trajectory = mclfi_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

  end function mclfi6_phi_trajectory

!returns sigmastar once phistar is computed
  function mclfi6_sigma_trajectory(phistar, Vparam, msigma, sigmaend)
    implicit none
    real(kp), intent(in) :: phistar, msigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: mclfi6_sigma_trajectory

    mclfi6_sigma_trajectory = mclfi_sigma_trajectory(phistar, Vparam, msigma, sigmaend)

  end function mclfi6_sigma_trajectory

!returns ns from already computed intermediary quantites (faciq)
  function mclfi6_ns_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mclfi6_ns_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mclfi_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mclfi6_ns_faciq = MC_ns(mclfi_potential,mclfi_deriv_potential,mclfi_deriv_second_potential, &
                mclfi_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mclfi6_ns_faciq

!returns r from already computed intermediary quantites (faciq)
  function mclfi6_r_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mclfi6_r_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mclfi_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mclfi6_r_faciq = MC_r(mclfi_potential,mclfi_deriv_potential,mclfi_deriv_second_potential, &
                Vparam,phistar,sigmastar,msigma,Delta)

  end function mclfi6_r_faciq

!returns fNL from already computed intermediary quantites (faciq)
  function mclfi6_fNL_faciq(msigma, Gammaphi, Gammasigma, sigmaend, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mclfi6_fNL_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mclfi_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mclfi6_fNL_faciq = MC6_fNL_phistarfree(mclfi_potential,mclfi_deriv_potential, &
               mclfi_deriv_second_potential, mclfi_phiend0, mclfi_InverseVp_primitive, &
               Vparam, msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mclfi6_fNL_faciq

!returns alphaS from already computed intermediary quantites (faciq)
  function mclfi6_alphaS_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mclfi6_alphaS_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mclfi_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mclfi6_alphaS_faciq = MC_alphaS(mclfi_potential,mclfi_deriv_potential,mclfi_deriv_second_potential, &
                mclfi_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mclfi6_alphaS_faciq



  end module mclfi6reheat
