!test the massive curvaton 6 - large field inflation functions
program mcni6main
  use infprec, only : kp, transfert
  use cosmopar, only : lnRhoNuc
  use infinout, only : delete_file, livewrite
  use mcni6reheat, only : mcni6_ns, mcni6_r, mcni6_fnl, mcni6_isnormalizable, mcni6_M, &
                           mcni6_bfoldstar, mcni6_Pzeta_sigma
  use mcnisr, only : mcni_potential,mcni_phiend0
  use curvaton, only : MC6_sigmaend_min, MC6_sigmaend_max

  implicit none

  integer :: i, j, k, l
  integer :: Nmsigma, NGammaphi, NGammasigma, Nsigmaend
  real(kp) ::  sigmaend, msigma, Gammaphi, Gammasigma, M, f, Psigma
  real(kp) ::  Gammaphimin, Gammaphimax, Gammasigmamin, Gammasigmamax, msigmamin, msigmamax, &
               sigmaendmin, sigmaendmax
  real(kp) :: ns, r, fnl, security, bfoldstar
  type(transfert) :: Vparam

  Nmsigma = 15
  NGammaphi = 15
  NGammasigma = 10
  Nsigmaend = 40

  security = 0.8_kp

  Gammasigmamin=sqrt(exp(lnRhoNuc)/3._kp)


  msigmamin = Gammasigmamin/security**9
  Gammaphimin = Gammasigmamin/security**9

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!                       !!!!
!!!!        f=3 Mp         !!!!
!!!!                       !!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  f=3._kp
  M = (7._kp*10**(-9.)*exp(-100._kp/f**2)/(1._kp+2._kp*f**2))**(0.25_kp) ! Fiducial Value, valid for f=O(1)
  msigmamax = M**2

  Vparam%real1 = M
  Vparam%real2 = f

  call delete_file('mcni6_feq3_predic.dat')

  do i=0,Nmsigma
    print*, '     ###################     '
    print*, '     ##               ##     '
    print*, '     ##               ##     '
    print*, '     ##  completed:   ##     '
    print*, '     ##  ',i,'/',Nmsigma,'         ##     '
    print*, '     ##               ##     '
    print*, '     ##               ##     '
    print*, '     ###################     '
    msigma=msigmamin*(msigmamax/msigmamin)**(real(i,kp)/real(Nmsigma,kp))
    !msigma=msigmamax*exp(log(msigmamin/msigmamax)*1._kp/(1._kp+10._kp*real(i,kp)/real(Nmsigma,kp))) !prefers sampling higher values
    Gammaphimax=msigma!*security
    do j=0,NGammaphi
      !Gammaphi=Gammaphimin*(Gammaphimax/Gammaphimin)**(real(j,kp)/real(NGammaphi,kp))
      Gammaphi=Gammaphimax*exp(log(Gammaphimin/Gammaphimax)*1._kp/(1._kp+10._kp*real(j,kp)/real(NGammaphi,kp))) !prefers sampling higher values
      Gammasigmamax=Gammaphi*security**9
      do k=0,NGammasigma
        !Gammasigma=Gammasigmamin*(Gammasigmamax/Gammasigmamin)**(real(j,kp)/real(NGammasigma,kp))
        Gammasigma=Gammasigmamax*exp(log(Gammasigmamin/Gammasigmamax)*1._kp/(1._kp+10._kp*real(k,kp)/real(NGammasigma,kp))) !prefers sampling higher values
        sigmaendmax = 20._kp ! 20 is to avoid wasting exploring useless parameters space (if sigmaend>20, impossible to normalize M)
        sigmaendmin=MC6_sigmaend_min(msigma,Gammaphi,Gammasigma)/security
        do l=0,Nsigmaend
          sigmaend=sigmaendmin*(sigmaendmax/sigmaendmin)**(real(l,kp)/real(Nsigmaend,kp))
          !sigmaend=sigmaendmax*exp(log(sigmaendmin/sigmaendmax)*1._kp/(1._kp+100._kp*real(l,kp)/real(Nsigmaend,kp))) !prefers sampling higher values
          if (mcni6_isnormalizable(f, msigma, Gammaphi, Gammasigma, sigmaend)) then
            M = mcni6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
            Vparam%real1 = M
            if (sigmaend < MC6_sigmaend_max(mcni_potential,mcni_phiend0,Vparam,msigma,Gammaphi,Gammasigma)) then ! check that sigmaend does not domimate the energy content at the end of inflation
              ns = mcni6_ns(f, msigma, Gammaphi, Gammasigma, sigmaend)
              r = mcni6_r(f, msigma, Gammaphi, Gammasigma, sigmaend)
              fnl = mcni6_fnl(f, msigma, Gammaphi, Gammasigma, sigmaend)
              Psigma = mcni6_Pzeta_sigma(f, msigma, Gammaphi, Gammasigma, sigmaend)
              bfoldstar = mcni6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
              print*, 'f=',f,'M = ',M, 'Gammaphi', Gammaphi,'msigma=', msigma, 'Gammasigma=', Gammasigma, &
                  'sigmaend', sigmaend
              print*, ' efoldstar=',-bfoldstar,' ns=',ns,' r=',r, 'fnl=',fnl, 'Psigma=',Psigma
              call livewrite('mcni6_feq3_predic.dat',M,msigma,Gammaphi,Gammasigma,sigmaend,ns,r,fnl,Psigma)
            end if
          end if
        end do
      end do
    end do
  end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!                       !!!!
!!!!        f=6 Mp         !!!!
!!!!                       !!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  f=6._kp
  M = (3.35_kp*10**(-9.)*f**2/50._kp**2)**0.25_kp ! Fiducial Value, valid for f>>1
  msigmamax = M**2

  Vparam%real1 = M
  Vparam%real2 = f

  call delete_file('mcni6_feq6_predic.dat')

  do i=0,Nmsigma
    print*, '     ###################     '
    print*, '     ##               ##     '
    print*, '     ##               ##     '
    print*, '     ##  completed:   ##     '
    print*, '     ##  ',i,'/',Nmsigma,'         ##     '
    print*, '     ##               ##     '
    print*, '     ##               ##     '
    print*, '     ###################     '
    msigma=msigmamin*(msigmamax/msigmamin)**(real(i,kp)/real(Nmsigma,kp))
    !msigma=msigmamax*exp(log(msigmamin/msigmamax)*1._kp/(1._kp+10._kp*real(i,kp)/real(Nmsigma,kp))) !prefers sampling higher values
    Gammaphimax=msigma!*security
    do j=0,NGammaphi
      !Gammaphi=Gammaphimin*(Gammaphimax/Gammaphimin)**(real(j,kp)/real(NGammaphi,kp))
      Gammaphi=Gammaphimax*exp(log(Gammaphimin/Gammaphimax)*1._kp/(1._kp+10._kp*real(j,kp)/real(NGammaphi,kp))) !prefers sampling higher values
      Gammasigmamax=Gammaphi*security**9
      do k=0,NGammasigma
        !Gammasigma=Gammasigmamin*(Gammasigmamax/Gammasigmamin)**(real(j,kp)/real(NGammasigma,kp))
        Gammasigma=Gammasigmamax*exp(log(Gammasigmamin/Gammasigmamax)*1._kp/(1._kp+10._kp*real(k,kp)/real(NGammasigma,kp))) !prefers sampling higher values
        sigmaendmax = 20._kp ! 20 is to avoid wasting exploring useless parameters space (if sigmaend>20, impossible to normalize M)
        sigmaendmin=MC6_sigmaend_min(msigma,Gammaphi,Gammasigma)/security
        do l=0,Nsigmaend
          sigmaend=sigmaendmin*(sigmaendmax/sigmaendmin)**(real(l,kp)/real(Nsigmaend,kp))
          !sigmaend=sigmaendmax*exp(log(sigmaendmin/sigmaendmax)*1._kp/(1._kp+100._kp*real(l,kp)/real(Nsigmaend,kp))) !prefers sampling higher values
          if (mcni6_isnormalizable(f, msigma, Gammaphi, Gammasigma, sigmaend)) then
            M = mcni6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
            Vparam%real1 = M
            if (sigmaend < MC6_sigmaend_max(mcni_potential,mcni_phiend0,Vparam,msigma,Gammaphi,Gammasigma)) then ! check that sigmaend does not domimate the energy content at the end of inflation
              ns = mcni6_ns(f, msigma, Gammaphi, Gammasigma, sigmaend)
              r = mcni6_r(f, msigma, Gammaphi, Gammasigma, sigmaend)
              fnl = mcni6_fnl(f, msigma, Gammaphi, Gammasigma, sigmaend)
              Psigma = mcni6_Pzeta_sigma(f, msigma, Gammaphi, Gammasigma, sigmaend)
              bfoldstar = mcni6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
              print*, 'f=',f,'M = ',M, 'Gammaphi', Gammaphi,'msigma=', msigma, 'Gammasigma=', Gammasigma, &
                  'sigmaend', sigmaend
              print*, ' efoldstar=',-bfoldstar,' ns=',ns,' r=',r, 'fnl=',fnl, 'Psigma=',Psigma
              call livewrite('mcni6_feq6_predic.dat',M,msigma,Gammaphi,Gammasigma,sigmaend,ns,r,fnl,Psigma)
            end if
          end if
        end do
      end do
    end do
  end do


end program mcni6main
