!reheating functions for the large field model with massive curvaton

  module mcni6reheat

    use infprec, only : kp, tolkp, transfert
    use inftools, only : zbrent
    use cosmopar, only : powerAmpScalar
    use mcnisr, only : mcni_potential, mcni_deriv_second_potential, mcni_phiend0, &
        mcni_phi_trajectory, mcni_deriv_potential, mcni_InverseVp_primitive, &
        mcni_phi_trajectory, mcni_deriv_third_potential, mcni_Delta_faciq, &
        mcni_sigma_trajectory
    use curvaton, only : find_MC_bfoldstar, MC6_lnRrad, MC6_Pzeta, MC_Delta, &
                         MC_inftraj_sigma, MC_r, MC_ns, MC6_fNL_phistarfree, &
                         MC6_Mmin_reh, MC6_Pzeta_relativecontribution_sigma, &
                         MC_alphaS, MC6_sigmaend_max

    implicit none

    private

    public mcni6_ns, mcni6_r, mcni6_fnl, mcni6_isnormalizable, mcni6_M, &
            mcni6_bfoldstar, mcni6_Pzeta_sigma, mcni6_alphaS, mcni6_sigma_trajectory, mcni6_ns_faciq, &
           mcni6_r_faciq, mcni6_fNL_faciq, mcni6_alphaS_faciq, mcni6_phi_trajectory

  contains


!!!!!!!!!!!!!!!!!!!!!!
!! Number of efolds !!
!!!!!!!!!!!!!!!!!!!!!!

!Returns the number of bfolds given potential parameters
  function mcni6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni6_bfoldstar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp), parameter :: tonind=tolkp
    real(kp) :: bfoldstarmini,bfoldstarmaxi,bfoldstar
    type(transfert) :: mcniData

    mcniData%real1 = Vparam%real1 ! M
    mcniData%real2 = Vparam%real2 ! f
    mcniData%real3 = msigma
    mcniData%real4 = Gammaphi
    mcniData%real5 = Gammasigma
    mcniData%real6 = sigmaend

    bfoldstarmini = -1000._kp
    bfoldstarmaxi = 0._kp

    bfoldstar = zbrent(find_mcni6_bfoldstar,bfoldstarmini,bfoldstarmaxi,tonind,mcniData)

    mcni6_bfoldstar = bfoldstar

  end function mcni6_bfoldstar

! Returns the equation one must solve to obtain bfoldstar from model parameters
  function find_mcni6_bfoldstar(bfoldstar,mcniData)
    implicit none
    real(kp) :: find_mcni6_bfoldstar
    real(kp), intent(in) :: bfoldstar
    type(transfert), optional, intent(inout) :: mcniData
    type(transfert) :: Vparam

    real(kp) :: msigma, Gammaphi, Gammasigma, sigmaend, lnRrad, Pstar, phistar

    Vparam%real1 = mcniData%real1! M
    Vparam%real2 = mcniData%real2 ! f

    msigma = mcniData%real3
    Gammaphi = mcniData%real4
    Gammasigma = mcniData%real5
    sigmaend = mcniData%real6

    lnRrad = MC6_lnRrad(mcni_potential, mcni_deriv_second_potential, mcni_phiend0, Vparam, &
             msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    Pstar = MC6_Pzeta(mcni_potential, mcni_deriv_potential, mcni_deriv_second_potential, &
            mcni_phiend0, mcni_InverseVp_primitive, Vparam, msigma, Gammaphi, Gammasigma, &
            sigmaend, phistar)

    find_mcni6_bfoldstar = find_MC_bfoldstar(mcni_potential, mcni_deriv_potential, &
         mcni_deriv_second_potential, mcni_phiend0, mcni_InverseVp_primitive, &
         mcni_phi_trajectory, Vparam, msigma, Gammaphi, Gammasigma, sigmaend, lnRrad, &
         Pstar, bfoldstar)

  end function find_mcni6_bfoldstar

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Power Spectrum Amplitude  !!
!!      Normalization         !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! Returns the power spectrum amplitude once the number of bfolds_star has been solved
  function mcni6_Pstar_bfoldssolved(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    real(kp) :: mcni6_Pstar_bfoldssolved
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: bfoldstar, phistar

    bfoldstar = mcni6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    mcni6_Pstar_bfoldssolved = MC6_Pzeta(mcni_potential, mcni_deriv_potential, &
        mcni_deriv_second_potential,mcni_phiend0, mcni_InverseVp_primitive, Vparam, msigma, &
        Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcni6_Pstar_bfoldssolved

! Returns the equation one must solve to obtain M from other model parameters by normalizing it to the scalar power spectrum amplitude
  function find_mcni6_M(M,mcniData)
    implicit none
    real(kp) :: find_mcni6_M
    real(kp), intent(in) :: M
    type(transfert), optional, intent(inout) :: mcniData
    real(kp) :: msigma, Gammaphi, Gammasigma, sigmaend, Pstar
    type(transfert) :: Vparam

    Vparam%real1 = M ! M
    Vparam%real2 = mcniData%real2 ! f

    msigma = mcniData%real3
    Gammaphi = mcniData%real4
    Gammasigma = mcniData%real5
    sigmaend = mcniData%real6

    Pstar = mcni6_Pstar_bfoldssolved(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)

    find_mcni6_M = Pstar - powerAmpScalar

  end function find_mcni6_M

!Returns the mass normalization of the inflaton potential (M) to the power spectrum amplitude
  function mcni6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni6_M
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp), parameter :: tonind=tolkp
    real(kp) :: Mmini1, Mmini2, Mmini3, Mmini, Mmaxi, M
    type(transfert) :: mcniData

    mcniData%real2 = Vparam%real2 ! f
    mcniData%real3 = msigma
    mcniData%real4 = Gammaphi
    mcniData%real5 = Gammasigma
    mcniData%real6 = sigmaend

    Vparam%real1 = 1._kp ! M (dummy value, to get V/M^4 in Mmini)

    Mmini1 = (msigma**2*sigmaend**2/(3._kp* &
              mcni_potential(mcni_phiend0(Vparam),Vparam)))**(0.25_kp)*100._kp ! This is to have phi dominate the energy budget at the end of inflation, 100 is a safety number
    Mmini2 = (msigma**2/ &
              mcni_potential(mcni_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have sigma still light at the end of inflation
    Mmini3 = MC6_Mmin_reh(mcni_potential, mcni_deriv_second_potential, mcni_phiend0, &
            Vparam, msigma, Gammaphi, Gammasigma, sigmaend)! CMB pivot scale exits Hubble radius during the first inflationary phase

    Mmini = max(Mmini1,Mmini2,Mmini3)*10._kp !10 is a safety number
    Mmaxi = 1._kp

    if (find_mcni6_M(Mmini,mcniData)*find_mcni6_M(Mmaxi,mcniData) > 0._kp) then
         stop 'normalization to Power Spectrum Amplitude is Impossible'
    endif

    if (Mmini > Mmaxi) then
         stop 'Mmini>Mmaxi'
    endif

    M = zbrent(find_mcni6_M,Mmini,Mmaxi,tonind,mcniData)

    mcni6_M = M

  end function mcni6_M

!Test Function to determine if the power spectrum is normalizable
  function mcni6_isnormalizable(f, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    logical :: mcni6_isnormalizable
    real(kp), intent(in) :: f, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam
    real(kp) :: Mmini1, Mmini2, Mmini3, Mmini, Mmaxi, M
    type(transfert) :: mcniData

    Vparam%real1 = 1._kp ! M (dummy value, to get V/M^4 in Mmini)
    Vparam%real2 = f

    mcni6_isnormalizable = .false.
    Mmini1 = (msigma**2*sigmaend**2/(3._kp* &
              mcni_potential(mcni_phiend0(Vparam),Vparam)))**(0.25_kp)*100._kp ! This is to have phi dominate the energy budget at the end of inflation, 100 is a safety number
    Mmini2 = (msigma**2/ &
              mcni_potential(mcni_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have sigma still light at the end of inflation
    Mmini3 = MC6_Mmin_reh(mcni_potential, mcni_deriv_second_potential, mcni_phiend0, Vparam, &
             msigma, Gammaphi, Gammasigma, sigmaend)! CMB pivot scale exits Hubble radius during the first inflationary phase
    Mmini = max(Mmini1,Mmini2,Mmini3)*10._kp !10 is a safety number
    Mmaxi = 1._kp
    mcniData%real2 = Vparam%real2 ! f
    mcniData%real3 = msigma
    mcniData%real4 = Gammaphi
    mcniData%real5 = Gammasigma
    mcniData%real6 = sigmaend

    if (Mmini<Mmaxi) then
      if (find_mcni6_M(Mmini,mcniData)*find_mcni6_M(Mmaxi,mcniData) < 0._kp) then
        M = mcni6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
        if (sigmaend < MC6_sigmaend_max(mcni_potential,mcni_phiend0,Vparam,msigma,Gammaphi,Gammasigma)) then ! check that sigmaend does not domimate the energy content at the end of inflation
          mcni6_isnormalizable = .true.
        endif
      endif
    endif

  end function mcni6_isnormalizable

  !Returns the correlation angle Delta between curvature and isocurvature power spectra
  function mcni6_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni6_Delta, mphi, bfoldstar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    type(transfert) :: Vparam_passed

    Vparam_passed%real1 = mcni6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) ! Normalizes M
    Vparam_passed%real2 = Vparam%real2 ! Passes f
    bfoldstar = mcni6_bfoldstar(Vparam_passed, msigma, Gammaphi, Gammasigma, sigmaend)
    mcni6_Delta = MC_Delta(mcni_potential, mcni_deriv_potential, mcni_deriv_second_potential, &
                 mcni_phiend0, mcni_InverseVp_primitive, mcni_phi_trajectory, Vparam_passed, &
                 msigma, Gammaphi, Gammasigma, sigmaend, bfoldstar)

  end function mcni6_Delta

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Observable Predictions   !!
!!    Direct Calculation     !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  function mcni6_ns(f, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni6_ns, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: f, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = f !passes f
    Vparam%real1 = mcni6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcni6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mcni_potential,mcni_deriv_second_potential,mcni_phiend0, &
                mcni_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mcni6_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mcni6_ns = MC_ns(mcni_potential,mcni_deriv_potential,mcni_deriv_second_potential, &
                mcni_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcni6_ns

  function mcni6_alphaS(f, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni6_alphaS, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: f, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = f !passes f
    Vparam%real1 = mcni6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcni6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mcni_potential,mcni_deriv_second_potential,mcni_phiend0, &
                mcni_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mcni6_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mcni6_alphaS = MC_alphaS(mcni_potential,mcni_deriv_potential,mcni_deriv_second_potential, &
                mcni_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcni6_alphaS

  function mcni6_r(f, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni6_r, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: f, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = f !passes f
    Vparam%real1 = mcni6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcni6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mcni_potential,mcni_deriv_second_potential,mcni_phiend0, &
                mcni_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mcni6_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mcni6_r = MC_r(mcni_potential,mcni_deriv_potential,mcni_deriv_second_potential,Vparam, &
               phistar,sigmastar,msigma,Delta)

  end function mcni6_r

  function mcni6_fNL(f, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni6_fNL, bfoldstar, phistar
    real(kp), intent(in) :: f, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = f !passes f
    Vparam%real1 = mcni6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcni6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

    mcni6_fNL=MC6_fNL_phistarfree(mcni_potential,mcni_deriv_potential, &
               mcni_deriv_second_potential, mcni_phiend0, mcni_InverseVp_primitive, &
               Vparam, msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcni6_fNL

! Returns the relative contribution of sigma to the total curvature power spectrum
  function mcni6_Pzeta_sigma(f, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni6_Pzeta_sigma, bfoldstar, phistar
    real(kp), intent(in) :: f, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = f !passes f
    Vparam%real1 = mcni6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcni6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

    mcni6_Pzeta_sigma=MC6_Pzeta_relativecontribution_sigma(mcni_potential,mcni_deriv_potential, &
                      mcni_deriv_second_potential,mcni_phiend0,mcni_InverseVp_primitive,Vparam, &
                      msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcni6_Pzeta_sigma

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Observable Predictions   !!
!! Step by Step Calculation  !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!returns phistar once bfoldstar is computed
  function mcni6_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    implicit none
    real(kp), intent(in) :: bfoldstar, msigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: mcni6_phi_trajectory

    mcni6_phi_trajectory = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

  end function mcni6_phi_trajectory

!returns sigmastar once phistar is computed
  function mcni6_sigma_trajectory(phistar, Vparam, msigma, sigmaend)
    implicit none
    real(kp), intent(in) :: phistar, msigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: mcni6_sigma_trajectory

    mcni6_sigma_trajectory = mcni_sigma_trajectory(phistar, Vparam, msigma, sigmaend)

  end function mcni6_sigma_trajectory

!returns ns from already computed intermediary quantites (faciq)
  function mcni6_ns_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcni6_ns_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcni_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcni6_ns_faciq = MC_ns(mcni_potential,mcni_deriv_potential,mcni_deriv_second_potential, &
                mcni_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcni6_ns_faciq

!returns r from already computed intermediary quantites (faciq)
  function mcni6_r_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcni6_r_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcni_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcni6_r_faciq = MC_r(mcni_potential,mcni_deriv_potential,mcni_deriv_second_potential, &
                Vparam,phistar,sigmastar,msigma,Delta)

  end function mcni6_r_faciq

!returns fNL from already computed intermediary quantites (faciq)
  function mcni6_fNL_faciq(msigma, Gammaphi, Gammasigma, sigmaend, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcni6_fNL_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcni_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcni6_fNL_faciq = MC6_fNL_phistarfree(mcni_potential,mcni_deriv_potential, &
               mcni_deriv_second_potential, mcni_phiend0, mcni_InverseVp_primitive, &
               Vparam, msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcni6_fNL_faciq

!returns alphaS from already computed intermediary quantites (faciq)
  function mcni6_alphaS_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcni6_alphaS_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcni_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcni6_alphaS_faciq = MC_alphaS(mcni_potential,mcni_deriv_potential,mcni_deriv_second_potential, &
                mcni_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcni6_alphaS_faciq



  end module mcni6reheat
