!reheating functions for the large field model with massive curvaton

  module mcni7reheat

    use infprec, only : kp, tolkp, transfert
    use inftools, only : zbrent
    use cosmopar, only : powerAmpScalar
    use mcnisr, only : mcni_potential, mcni_deriv_second_potential, mcni_phiend0, &
        mcni_phi_trajectory, mcni_deriv_potential, mcni_InverseVp_primitive, &
        mcni_phi_trajectory, mcni_deriv_third_potential, mcni_Delta_faciq, &
        mcni_sigma_trajectory
    use curvaton, only : find_MC_bfoldstar, MC7_lnRrad, MC7_Pzeta, MC_Delta, &
                         MC_inftraj_sigma, MC_r, MC_ns, MC7_fNL_phistarfree, &
                         MC7_Pzeta_relativecontribution_sigma, MC_alphaS

    implicit none

    private

    public mcni7_ns, mcni7_r, mcni7_fnl, mcni7_isnormalizable, mcni7_M, &
            mcni7_bfoldstar, mcni7_Pzeta_sigma, mcni7_alphaS, mcni7_sigma_trajectory, mcni7_ns_faciq, &
           mcni7_r_faciq, mcni7_fNL_faciq, mcni7_alphaS_faciq, mcni7_phi_trajectory

  contains


!!!!!!!!!!!!!!!!!!!!!!
!! Number of efolds !!
!!!!!!!!!!!!!!!!!!!!!!

!Returns the number of bfolds given potential parameters
  function mcni7_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni7_bfoldstar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp), parameter :: tonind=tolkp
    real(kp) :: bfoldstarmini,bfoldstarmaxi,bfoldstar
    type(transfert) :: mcniData

    mcniData%real1 = Vparam%real1 ! M
    mcniData%real2 = Vparam%real2 ! f
    mcniData%real3 = msigma
    mcniData%real4 = Gammaphi
    mcniData%real5 = Gammasigma
    mcniData%real6 = sigmaend

    bfoldstarmini = -1000._kp
    bfoldstarmaxi = 0._kp

    bfoldstar = zbrent(find_mcni7_bfoldstar,bfoldstarmini,bfoldstarmaxi,tonind,mcniData)

    mcni7_bfoldstar = bfoldstar

  end function mcni7_bfoldstar

! Returns the equation one must solve to obtain bfoldstar from model parameters
  function find_mcni7_bfoldstar(bfoldstar,mcniData)
    implicit none
    real(kp) :: find_mcni7_bfoldstar
    real(kp), intent(in) :: bfoldstar
    type(transfert), optional, intent(inout) :: mcniData
    type(transfert) :: Vparam

    real(kp) :: msigma, Gammaphi, Gammasigma, sigmaend, lnRrad, Pstar, phistar

    Vparam%real1 = mcniData%real1! M
    Vparam%real2 = mcniData%real2 ! f

    msigma = mcniData%real3
    Gammaphi = mcniData%real4
    Gammasigma = mcniData%real5
    sigmaend = mcniData%real6

    lnRrad = MC7_lnRrad(mcni_potential, mcni_deriv_second_potential, mcni_phiend0, Vparam, &
             msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    Pstar = MC7_Pzeta(mcni_potential, mcni_deriv_potential, mcni_deriv_second_potential, &
            mcni_phiend0, mcni_InverseVp_primitive, Vparam, msigma, Gammaphi, Gammasigma, &
            sigmaend, phistar)

    find_mcni7_bfoldstar = find_MC_bfoldstar(mcni_potential, mcni_deriv_potential, &
         mcni_deriv_second_potential, mcni_phiend0, mcni_InverseVp_primitive, &
         mcni_phi_trajectory, Vparam, msigma, Gammaphi, Gammasigma, sigmaend, lnRrad, &
         Pstar, bfoldstar)

  end function find_mcni7_bfoldstar

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Power Spectrum Amplitude  !!
!!      Normalization         !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! Returns the power spectrum amplitude once the number of bfolds_star has been solved
  function mcni7_Pstar_bfoldssolved(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    real(kp) :: mcni7_Pstar_bfoldssolved
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: bfoldstar, phistar

    bfoldstar = mcni7_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    mcni7_Pstar_bfoldssolved = MC7_Pzeta(mcni_potential, mcni_deriv_potential, &
        mcni_deriv_second_potential,mcni_phiend0, mcni_InverseVp_primitive, Vparam, msigma, &
        Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcni7_Pstar_bfoldssolved

! Returns the equation one must solve to obtain M from other model parameters by normalizing it to the scalar power spectrum amplitude
  function find_mcni7_M(M,mcniData)
    implicit none
    real(kp) :: find_mcni7_M
    real(kp), intent(in) :: M
    type(transfert), optional, intent(inout) :: mcniData
    real(kp) :: msigma, Gammaphi, Gammasigma, sigmaend, Pstar
    type(transfert) :: Vparam

    Vparam%real1 = M ! M
    Vparam%real2 = mcniData%real2 ! f

    msigma = mcniData%real3
    Gammaphi = mcniData%real4
    Gammasigma = mcniData%real5
    sigmaend = mcniData%real6

    Pstar = mcni7_Pstar_bfoldssolved(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)

    find_mcni7_M = Pstar - powerAmpScalar

  end function find_mcni7_M

!Returns the mass normalization of the inflaton potential (M) to the power spectrum amplitude
  function mcni7_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni7_M
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp), parameter :: tonind=tolkp
    real(kp) :: Mmini1, Mmini2, Mmini3, Mmini, Mmaxi, M
    type(transfert) :: mcniData

    mcniData%real2 = Vparam%real2 ! f
    mcniData%real3 = msigma
    mcniData%real4 = Gammaphi
    mcniData%real5 = Gammasigma
    mcniData%real6 = sigmaend

    Vparam%real1 = 1._kp ! M (dummy value, to get V/M^4 in Mmini)

    Mmini1 = (msigma**2*sigmaend**2/(3._kp* &
              mcni_potential(mcni_phiend0(Vparam),Vparam)))**(0.25_kp) ! This is to have phi dominate the energy budget at the end of inflation
    Mmini2 = (msigma**2/ &
              mcni_potential(mcni_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have sigma still light at the end of inflation
    Mmini3 = (2._kp*Gammaphi**2/ mcni_potential(mcni_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have Hend>Gammaphi
    Mmini = max(Mmini1,Mmini2,Mmini3)*1.1_kp !1.1 is a safety number
    Mmaxi = 1._kp

    if (find_mcni7_M(Mmini,mcniData)*find_mcni7_M(Mmaxi,mcniData) > 0._kp) then
         stop 'normalization to Power Spectrum Amplitude is Impossible'
    endif

    M = zbrent(find_mcni7_M,Mmini,Mmaxi,tonind,mcniData)

    mcni7_M = M

  end function mcni7_M

  !Returns the correlation angle Delta between curvature and isocurvature power spectra
  function mcni7_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni7_Delta, mphi, bfoldstar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    type(transfert) :: Vparam_passed

    Vparam_passed%real1 = mcni7_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) ! Normalizes M
    Vparam_passed%real2 = Vparam%real2 ! passes f
    bfoldstar = mcni7_bfoldstar(Vparam_passed, msigma, Gammaphi, Gammasigma, sigmaend)
    mcni7_Delta = MC_Delta(mcni_potential, mcni_deriv_potential, &
            mcni_deriv_second_potential, mcni_phiend0, mcni_InverseVp_primitive, &
            mcni_phi_trajectory, Vparam_passed, msigma, Gammaphi, Gammasigma, &
            sigmaend, bfoldstar)

  end function mcni7_Delta

!Test Function to determine if the power spectrum is normalizable
  function mcni7_isnormalizable(f, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    logical :: mcni7_isnormalizable
    real(kp), intent(in) :: f, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam
    real(kp) :: Mmini1, Mmini2, Mmini3, Mmini, Mmaxi
    type(transfert) :: mcniData

    Vparam%real1 = 1._kp ! M (dummy value, to get V/M^4 in Mmini)
    Vparam%real2 = f

    mcni7_isnormalizable = .false.
    Mmini1 = (msigma**2*sigmaend**2/(3._kp* &
              mcni_potential(mcni_phiend0(Vparam),Vparam)))**(0.25_kp) ! This is to have phi dominate the energy budget at the end of inflation
    Mmini2 = (msigma**2/ &
              mcni_potential(mcni_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have sigma still light at the end of inflation
    Mmini3 = (2._kp*Gammaphi**2/ mcni_potential(mcni_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have Hend>Gammaphi
    Mmini = max(Mmini1,Mmini2,Mmini3)*1.1_kp !1.1 is a safety number
    Mmaxi = 1._kp
    mcniData%real2 = Vparam%real2 ! f
    mcniData%real3 = msigma
    mcniData%real4 = Gammaphi
    mcniData%real5 = Gammasigma
    mcniData%real6 = sigmaend

    if (find_mcni7_M(Mmini,mcniData)*find_mcni7_M(Mmaxi,mcniData) < 0._kp) then
      mcni7_isnormalizable = .true.
    endif

  end function mcni7_isnormalizable

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Observable Predictions   !!
!!    Direct Calculation     !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  function mcni7_ns(f, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni7_ns, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: f, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = f !passes f
    Vparam%real1 = mcni7_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcni7_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mcni_potential,mcni_deriv_second_potential,mcni_phiend0, &
                mcni_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mcni7_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mcni7_ns = MC_ns(mcni_potential,mcni_deriv_potential,mcni_deriv_second_potential, &
                mcni_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcni7_ns

  function mcni7_alphaS(f, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni7_alphaS, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: f, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = f !passes f
    Vparam%real1 = mcni7_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcni7_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mcni_potential,mcni_deriv_second_potential,mcni_phiend0, &
                mcni_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mcni7_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mcni7_alphaS = MC_alphaS(mcni_potential,mcni_deriv_potential,mcni_deriv_second_potential, &
                mcni_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcni7_alphaS

  function mcni7_r(f, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni7_r, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: f, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = f !passes f
    Vparam%real1 = mcni7_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcni7_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mcni_potential,mcni_deriv_second_potential,mcni_phiend0, &
                mcni_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mcni7_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mcni7_r = MC_r(mcni_potential,mcni_deriv_potential,mcni_deriv_second_potential, &
                Vparam,phistar,sigmastar,msigma,Delta)

  end function mcni7_r

  function mcni7_fNL(f, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni7_fNL, bfoldstar, phistar
    real(kp), intent(in) :: f, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = f !passes f
    Vparam%real1 = mcni7_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcni7_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

    mcni7_fNL=MC7_fNL_phistarfree(mcni_potential,mcni_deriv_potential, &
               mcni_deriv_second_potential, mcni_phiend0, mcni_InverseVp_primitive, &
               Vparam, msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcni7_fNL

! Returns the relative contribution of sigma to the total curvature power spectrum
  function mcni7_Pzeta_sigma(f, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcni7_Pzeta_sigma, bfoldstar, phistar
    real(kp), intent(in) :: f, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = f !passes f
    Vparam%real1 = mcni7_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcni7_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

    mcni7_Pzeta_sigma=MC7_Pzeta_relativecontribution_sigma(mcni_potential,mcni_deriv_potential, &
                      mcni_deriv_second_potential,mcni_phiend0,mcni_InverseVp_primitive,Vparam, &
                      msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcni7_Pzeta_sigma

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Observable Predictions   !!
!! Step by Step Calculation  !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!returns phistar once bfoldstar is computed
  function mcni7_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    implicit none
    real(kp), intent(in) :: bfoldstar, msigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: mcni7_phi_trajectory

    mcni7_phi_trajectory = mcni_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

  end function mcni7_phi_trajectory

!returns sigmastar once phistar is computed
  function mcni7_sigma_trajectory(phistar, Vparam, msigma, sigmaend)
    implicit none
    real(kp), intent(in) :: phistar, msigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: mcni7_sigma_trajectory

    mcni7_sigma_trajectory = mcni_sigma_trajectory(phistar, Vparam, msigma, sigmaend)

  end function mcni7_sigma_trajectory

!returns ns from already computed intermediary quantites (faciq)
  function mcni7_ns_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcni7_ns_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcni_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcni7_ns_faciq = MC_ns(mcni_potential,mcni_deriv_potential,mcni_deriv_second_potential, &
                mcni_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcni7_ns_faciq

!returns r from already computed intermediary quantites (faciq)
  function mcni7_r_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcni7_r_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcni_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcni7_r_faciq = MC_r(mcni_potential,mcni_deriv_potential,mcni_deriv_second_potential, &
                Vparam,phistar,sigmastar,msigma,Delta)

  end function mcni7_r_faciq

!returns fNL from already computed intermediary quantites (faciq)
  function mcni7_fNL_faciq(msigma, Gammaphi, Gammasigma, sigmaend, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcni7_fNL_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcni_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcni7_fNL_faciq = MC7_fNL_phistarfree(mcni_potential,mcni_deriv_potential, &
               mcni_deriv_second_potential, mcni_phiend0, mcni_InverseVp_primitive, &
               Vparam, msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcni7_fNL_faciq

!returns alphaS from already computed intermediary quantites (faciq)
  function mcni7_alphaS_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcni7_alphaS_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcni_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcni7_alphaS_faciq = MC_alphaS(mcni_potential,mcni_deriv_potential,mcni_deriv_second_potential, &
                mcni_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcni7_alphaS_faciq



  end module mcni7reheat
