!reheating functions for the power law model with massive curvaton

  module mcpli10reheat

    use infprec, only : kp, tolkp, transfert
    use inftools, only : zbrent
    use cosmopar, only : powerAmpScalar
    use mcplisr, only : mcpli_potential, mcpli_deriv_second_potential, mcpli_phiend0, &
        mcpli_phi_trajectory, mcpli_deriv_potential, mcpli_InverseVp_primitive, &
        mcpli_phi_trajectory, mcpli_deriv_third_potential, mcpli_Delta_faciq, &
        mcpli_sigma_trajectory
    use curvaton, only : find_MC_bfoldstar, MC10_lnRrad, MC10_Pzeta, MC_Delta, &
                        MC_inftraj_sigma, MC_r, MC_ns, MC10_fNL_phistarfree, &
                        MC10_Mmin_reh, MC10_Pzeta_relativecontribution_sigma, &
                        MC_alphaS, MC10_sigmaend_max

    implicit none

    private

    public mcpli10_ns, mcpli10_r, mcpli10_fnl, mcpli10_isnormalizable, mcpli10_M, &
            mcpli10_bfoldstar, mcpli10_Pzeta_sigma, mcpli10_alphaS, mcpli10_sigma_trajectory, mcpli10_ns_faciq, &
           mcpli10_r_faciq, mcpli10_fNL_faciq, mcpli10_alphaS_faciq, mcpli10_phi_trajectory

  contains


!!!!!!!!!!!!!!!!!!!!!!
!! Number of efolds !!
!!!!!!!!!!!!!!!!!!!!!!

!Returns the number of bfolds given potential parameters
  function mcpli10_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli10_bfoldstar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp), parameter :: tonind=tolkp
    real(kp) :: bfoldstarmini,bfoldstarmaxi,bfoldstar
    type(transfert) :: mcpliData

    mcpliData%real1 = Vparam%real1 ! M
    mcpliData%real2 = Vparam%real2 ! alpha
    mcpliData%real3 = msigma
    mcpliData%real4 = Gammaphi
    mcpliData%real5 = Gammasigma
    mcpliData%real6 = sigmaend

    bfoldstarmini = -1000._kp
    bfoldstarmaxi = 0._kp

    bfoldstar = zbrent(find_mcpli10_bfoldstar,bfoldstarmini,bfoldstarmaxi,tonind,mcpliData)

    mcpli10_bfoldstar = bfoldstar

  end function mcpli10_bfoldstar

! Returns the equation one must solve to obtain bfoldstar from model parameters
  function find_mcpli10_bfoldstar(bfoldstar,mcpliData)
    implicit none
    real(kp) :: find_mcpli10_bfoldstar
    real(kp), intent(in) :: bfoldstar
    type(transfert), optional, intent(inout) :: mcpliData
    type(transfert) :: Vparam

    real(kp) :: msigma, Gammaphi, Gammasigma, sigmaend, lnRrad, Pstar, phistar

    Vparam%real1 = mcpliData%real1! M
    Vparam%real2 = mcpliData%real2 ! alpha

    msigma = mcpliData%real3
    Gammaphi = mcpliData%real4
    Gammasigma = mcpliData%real5
    sigmaend = mcpliData%real6

    lnRrad = MC10_lnRrad(mcpli_potential, mcpli_deriv_second_potential, mcpli_phiend0, Vparam, &
             msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    Pstar = MC10_Pzeta(mcpli_potential, mcpli_deriv_potential, mcpli_deriv_second_potential, &
            mcpli_phiend0, mcpli_InverseVp_primitive, Vparam, msigma, Gammaphi, Gammasigma, &
            sigmaend, phistar)

    find_mcpli10_bfoldstar = find_MC_bfoldstar(mcpli_potential, mcpli_deriv_potential, &
         mcpli_deriv_second_potential, mcpli_phiend0, mcpli_InverseVp_primitive, &
         mcpli_phi_trajectory, Vparam, msigma, Gammaphi, Gammasigma, sigmaend, lnRrad, &
         Pstar, bfoldstar)

  end function find_mcpli10_bfoldstar

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Power Spectrum Amplitude  !!
!!      Normalization         !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! Returns the power spectrum amplitude once the number of bfolds_star has been solved
  function mcpli10_Pstar_bfoldssolved(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    real(kp) :: mcpli10_Pstar_bfoldssolved
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: bfoldstar, phistar

    bfoldstar = mcpli10_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    mcpli10_Pstar_bfoldssolved = MC10_Pzeta(mcpli_potential, mcpli_deriv_potential, &
        mcpli_deriv_second_potential,mcpli_phiend0, mcpli_InverseVp_primitive, Vparam, msigma, &
        Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcpli10_Pstar_bfoldssolved

! Returns the equation one must solve to obtain M from other model parameters by normalizing it to the scalar power spectrum amplitude
  function find_mcpli10_M(M,mcpliData)
    implicit none
    real(kp) :: find_mcpli10_M
    real(kp), intent(in) :: M
    type(transfert), optional, intent(inout) :: mcpliData
    real(kp) :: msigma, Gammaphi, Gammasigma, sigmaend, Pstar
    type(transfert) :: Vparam

    Vparam%real1 = M ! M
    Vparam%real2 = mcpliData%real2 ! alpha

    msigma = mcpliData%real3
    Gammaphi = mcpliData%real4
    Gammasigma = mcpliData%real5
    sigmaend = mcpliData%real6

    Pstar = mcpli10_Pstar_bfoldssolved(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)

    find_mcpli10_M = Pstar - powerAmpScalar

  end function find_mcpli10_M

!Returns the mass normalization of the inflaton potential (M) to the power spectrum amplitude
  function mcpli10_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli10_M
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp), parameter :: tonind=tolkp
    real(kp) :: Mmini1, Mmini2, Mmini3, Mmini4, Mmini, Mmaxi, M
    type(transfert) :: mcpliData

    mcpliData%real2 = Vparam%real2 ! alpha
    mcpliData%real3 = msigma
    mcpliData%real4 = Gammaphi
    mcpliData%real5 = Gammasigma
    mcpliData%real6 = sigmaend

    Vparam%real1 = 1._kp ! M (dummy value, to get V/M^4 in Mmini)

    Mmini1 = (msigma**2*sigmaend**2/(3._kp* &
              mcpli_potential(mcpli_phiend0(Vparam),Vparam)))**(0.25_kp) ! This is to have phi dominate the energy budget at the end of inflation
    Mmini2 = (msigma**2/ &
              mcpli_potential(mcpli_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have sigma still light at the end of inflation
    Mmini3 = (2._kp*Gammaphi**2/ mcpli_potential(mcpli_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have Hend>Gammaphi
    Mmini4 = MC10_Mmin_reh(mcpli_potential, mcpli_deriv_second_potential, mcpli_phiend0, &
            Vparam, msigma, Gammaphi, Gammasigma, sigmaend)! CMB pivot scale exits Hubble radius during the first inflationary phase

    Mmini = max(Mmini1,Mmini2,Mmini3,Mmini4)*10._kp !10. is a safety number
    Mmaxi = 100._kp

    if (find_mcpli10_M(Mmini,mcpliData)*find_mcpli10_M(Mmaxi,mcpliData) > 0._kp) then
         stop 'normalization to Power Spectrum Amplitude is Impossible'
    endif

    if (Mmini > Mmaxi) then
         stop 'Mmini>Mmaxi'
    endif

    M = zbrent(find_mcpli10_M,Mmini,Mmaxi,tonind,mcpliData)

    mcpli10_M = M

  end function mcpli10_M

!Test Function to determine if the power spectrum is normalizable
  function mcpli10_isnormalizable(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    logical :: mcpli10_isnormalizable
    real(kp), intent(in) :: alpha, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam
    real(kp) :: Mmini1, Mmini2, Mmini3, Mmini4, Mmini, Mmaxi, M
    type(transfert) :: mcpliData

    Vparam%real1 = 1._kp ! M (dummy value, to get V/M^4 in Mmini)
    Vparam%real2 = alpha

    mcpli10_isnormalizable = .false.
    Mmini1 = (msigma**2*sigmaend**2/(3._kp* &
              mcpli_potential(mcpli_phiend0(Vparam),Vparam)))**(0.25_kp) ! This is to have phi dominate the energy budget at the end of inflation
    Mmini2 = (msigma**2/ &
              mcpli_potential(mcpli_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have sigma still light at the end of inflation
    Mmini3 = (2._kp*Gammaphi**2/ mcpli_potential(mcpli_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have Hend>Gammaphi
    Mmini4 = MC10_Mmin_reh(mcpli_potential, mcpli_deriv_second_potential, mcpli_phiend0, &
            Vparam, msigma, Gammaphi, Gammasigma, sigmaend)! CMB pivot scale exits Hubble radius during the first inflationary phase
    Mmini = max(Mmini1,Mmini2,Mmini3,Mmini4)*10._kp !10. is a safety number
    Mmaxi = 100._kp
    mcpliData%real2 = Vparam%real2 ! alpha
    mcpliData%real3 = msigma
    mcpliData%real4 = Gammaphi
    mcpliData%real5 = Gammasigma
    mcpliData%real6 = sigmaend

    if (Mmini<Mmaxi) then
      if (find_mcpli10_M(Mmini,mcpliData)*find_mcpli10_M(Mmaxi,mcpliData) < 0._kp) then
        M = mcpli10_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
          if (sigmaend < MC10_sigmaend_max(mcpli_potential,mcpli_phiend0,&
                Vparam,msigma,Gammaphi,Gammasigma)) then ! check that sigmaend does not domimate the energy content at the end of inflation
          mcpli10_isnormalizable = .true.
        endif
      endif
    endif

  end function mcpli10_isnormalizable

  !Returns the correlation angle Delta between curvature and isocurvature power spectra
  function mcpli10_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli10_Delta, mphi, bfoldstar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    type(transfert) :: Vparam_passed

    Vparam_passed%real1 = mcpli10_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) ! Normalizes M
    Vparam_passed%real2 = Vparam%real2 ! Passes alpha
    bfoldstar = mcpli10_bfoldstar(Vparam_passed, msigma, Gammaphi, Gammasigma, sigmaend)
    mcpli10_Delta = MC_Delta(mcpli_potential, mcpli_deriv_potential, mcpli_deriv_second_potential, &
                 mcpli_phiend0, mcpli_InverseVp_primitive, mcpli_phi_trajectory, Vparam_passed, &
                 msigma, Gammaphi, Gammasigma, sigmaend, bfoldstar)

  end function mcpli10_Delta

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Observable Predictions   !!
!!    Direct Calculation     !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  function mcpli10_ns(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli10_ns, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: alpha, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = alpha !passes alpha
    Vparam%real1 = mcpli10_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcpli10_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mcpli_potential,mcpli_deriv_second_potential,mcpli_phiend0, &
                mcpli_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mcpli10_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mcpli10_ns = MC_ns(mcpli_potential,mcpli_deriv_potential,mcpli_deriv_second_potential, &
                mcpli_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcpli10_ns

  function mcpli10_alphaS(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli10_alphaS, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: alpha, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = alpha !passes alpha
    Vparam%real1 = mcpli10_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcpli10_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mcpli_potential,mcpli_deriv_second_potential,mcpli_phiend0, &
                mcpli_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mcpli10_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mcpli10_alphaS = MC_alphaS(mcpli_potential,mcpli_deriv_potential,mcpli_deriv_second_potential, &
                mcpli_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcpli10_alphaS

  function mcpli10_r(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli10_r, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: alpha, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = alpha !passes alpha
    Vparam%real1 = mcpli10_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcpli10_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mcpli_potential,mcpli_deriv_second_potential,mcpli_phiend0, &
                mcpli_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mcpli10_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mcpli10_r = MC_r(mcpli_potential,mcpli_deriv_potential,mcpli_deriv_second_potential,Vparam, &
               phistar,sigmastar,msigma,Delta)

  end function mcpli10_r

  function mcpli10_fNL(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli10_fNL, bfoldstar, phistar
    real(kp), intent(in) :: alpha, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = alpha !passes alpha
    Vparam%real1 = mcpli10_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcpli10_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

    mcpli10_fNL=MC10_fNL_phistarfree(mcpli_potential,mcpli_deriv_potential, &
               mcpli_deriv_second_potential, mcpli_phiend0, mcpli_InverseVp_primitive, &
               Vparam, msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcpli10_fNL

! Returns the relative contribution of sigma to the total curvature power spectrum
  function mcpli10_Pzeta_sigma(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli10_Pzeta_sigma, bfoldstar, phistar
    real(kp), intent(in) :: alpha, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = alpha !passes alpha
    Vparam%real1 = mcpli10_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcpli10_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

    mcpli10_Pzeta_sigma=MC10_Pzeta_relativecontribution_sigma(mcpli_potential,mcpli_deriv_potential, &
                      mcpli_deriv_second_potential,mcpli_phiend0,mcpli_InverseVp_primitive,Vparam, &
                      msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcpli10_Pzeta_sigma

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Observable Predictions   !!
!! Step by Step Calculation  !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!returns phistar once bfoldstar is computed
  function mcpli10_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    implicit none
    real(kp), intent(in) :: bfoldstar, msigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: mcpli10_phi_trajectory

    mcpli10_phi_trajectory = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

  end function mcpli10_phi_trajectory

!returns sigmastar once phistar is computed
  function mcpli10_sigma_trajectory(phistar, Vparam, msigma, sigmaend)
    implicit none
    real(kp), intent(in) :: phistar, msigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: mcpli10_sigma_trajectory

    mcpli10_sigma_trajectory = mcpli_sigma_trajectory(phistar, Vparam, msigma, sigmaend)

  end function mcpli10_sigma_trajectory

!returns ns from already computed intermediary quantites (faciq)
  function mcpli10_ns_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcpli10_ns_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcpli_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcpli10_ns_faciq = MC_ns(mcpli_potential,mcpli_deriv_potential,mcpli_deriv_second_potential, &
                mcpli_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcpli10_ns_faciq

!returns r from already computed intermediary quantites (faciq)
  function mcpli10_r_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcpli10_r_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcpli_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcpli10_r_faciq = MC_r(mcpli_potential,mcpli_deriv_potential,mcpli_deriv_second_potential, &
                Vparam,phistar,sigmastar,msigma,Delta)

  end function mcpli10_r_faciq

!returns fNL from already computed intermediary quantites (faciq)
  function mcpli10_fNL_faciq(msigma, Gammaphi, Gammasigma, sigmaend, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcpli10_fNL_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcpli_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcpli10_fNL_faciq = MC10_fNL_phistarfree(mcpli_potential,mcpli_deriv_potential, &
               mcpli_deriv_second_potential, mcpli_phiend0, mcpli_InverseVp_primitive, &
               Vparam, msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcpli10_fNL_faciq

!returns alphaS from already computed intermediary quantites (faciq)
  function mcpli10_alphaS_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcpli10_alphaS_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcpli_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcpli10_alphaS_faciq = MC_alphaS(mcpli_potential,mcpli_deriv_potential,mcpli_deriv_second_potential, &
                mcpli_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcpli10_alphaS_faciq





  end module mcpli10reheat
