!test the massive curvaton 6 - power law inflation functions
program mcpli6main
  use infprec, only : kp, transfert
  use cosmopar, only : lnRhoNuc
  use infinout, only : delete_file, livewrite
  use mcpli6reheat, only : mcpli6_ns, mcpli6_r, mcpli6_fnl, mcpli6_isnormalizable, mcpli6_M, &
                           mcpli6_bfoldstar, mcpli6_Pzeta_sigma
  use mcplisr, only : mcpli_potential,mcpli_phiend0
  use curvaton, only : MC6_sigmaend_min, MC6_sigmaend_max

  implicit none

  integer :: i, j, k, l
  integer :: Nmsigma, NGammaphi, NGammasigma, Nsigmaend
  real(kp) ::  sigmaend, msigma, Gammaphi, Gammasigma, M, alpha, Psigma
  real(kp) ::  Gammaphimin, Gammaphimax, Gammasigmamin, Gammasigmamax, msigmamin, msigmamax, &
               sigmaendmin, sigmaendmax
  real(kp) :: ns, r, fnl, security, bfoldstar
  type(transfert) :: Vparam

  real(kp), parameter :: junkphiend = 100._kp

  Nmsigma = 15
  NGammaphi = 15
  NGammasigma = 10
  Nsigmaend = 40

  security = 0.8_kp

  Gammasigmamin=sqrt(exp(lnRhoNuc)/3._kp)


  msigmamin = Gammasigmamin/security**9
  Gammaphimin = Gammasigmamin/security**9

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!                       !!!!
!!!!      alpha=0.02       !!!!
!!!!                       !!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  alpha=0.02_kp
  M = (720._kp*acos(-1._kp)**2*alpha**2*exp(-alpha**2*50._kp)*2._kp*10._kp**(-9.)/60._kp*exp(alpha*junkphiend))**(0.25_kp)! Fiducial Value
  msigmamax = M**2

  Vparam%real1 = M
  Vparam%real2 = alpha

  call delete_file('mcpli6_alphaeqpoint02_predic.dat')

  do i=0,Nmsigma
    print*, '     ###################     '
    print*, '     ##               ##     '
    print*, '     ##               ##     '
    print*, '     ##  completed:   ##     '
    print*, '     ##  ',i,'/',Nmsigma,'         ##     '
    print*, '     ##               ##     '
    print*, '     ##               ##     '
    print*, '     ###################     '
    msigma=msigmamin*(msigmamax/msigmamin)**(real(i,kp)/real(Nmsigma,kp))
    !msigma=msigmamax*exp(log(msigmamin/msigmamax)*1._kp/(1._kp+10._kp*real(i,kp)/real(Nmsigma,kp))) !prefers sampling higher values
    Gammaphimax=msigma!*security
    do j=0,NGammaphi
      !Gammaphi=Gammaphimin*(Gammaphimax/Gammaphimin)**(real(j,kp)/real(NGammaphi,kp))
      Gammaphi=Gammaphimax*exp(log(Gammaphimin/Gammaphimax)*1._kp/(1._kp+10._kp*real(j,kp)/real(NGammaphi,kp))) !prefers sampling higher values
      Gammasigmamax=Gammaphi*security**9
      do k=0,NGammasigma
        !Gammasigma=Gammasigmamin*(Gammasigmamax/Gammasigmamin)**(real(j,kp)/real(NGammasigma,kp))
        Gammasigma=Gammasigmamax*exp(log(Gammasigmamin/Gammasigmamax)*1._kp/(1._kp+10._kp*real(k,kp)/real(NGammasigma,kp))) !prefers sampling higher values
        sigmaendmax = 20._kp ! 20 is to avoid wasting exploring useless parameters space (if sigmaend>20, impossible to normalize M)
        sigmaendmin=MC6_sigmaend_min(msigma,Gammaphi,Gammasigma)/security
        do l=0,Nsigmaend
          sigmaend=sigmaendmin*(sigmaendmax/sigmaendmin)**(real(l,kp)/real(Nsigmaend,kp))
          !sigmaend=sigmaendmax*exp(log(sigmaendmin/sigmaendmax)*1._kp/(1._kp+100._kp*real(l,kp)/real(Nsigmaend,kp))) !prefers sampling higher values
          if (mcpli6_isnormalizable(alpha, msigma, Gammaphi, Gammasigma, sigmaend)) then
            M = mcpli6_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
            Vparam%real1 = M
            if (sigmaend < MC6_sigmaend_max(mcpli_potential,mcpli_phiend0,Vparam,msigma,Gammaphi,Gammasigma)) then ! check that sigmaend does not domimate the energy content at the end of inflation
              ns = mcpli6_ns(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
              r = mcpli6_r(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
              fnl = mcpli6_fnl(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
              Psigma = mcpli6_Pzeta_sigma(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
              bfoldstar = mcpli6_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
              print*, 'alpha=',alpha,'M = ',M, 'Gammaphi', Gammaphi,'msigma=', msigma, 'Gammasigma=', Gammasigma, &
                  'sigmaend', sigmaend
              print*, ' efoldstar=',-bfoldstar,' ns=',ns,' r=',r, 'fnl=',fnl, 'Psigma=',Psigma
              call livewrite('mcpli6_alphaeqpoint02_predic.dat',M,msigma,Gammaphi,Gammasigma,sigmaend,ns,r,fnl,Psigma)
            end if
          end if
        end do
      end do
    end do
  end do


end program mcpli6main
