!reheating functions for the power law model with massive curvaton

  module mcpli7reheat

    use infprec, only : kp, tolkp, transfert
    use inftools, only : zbrent
    use cosmopar, only : powerAmpScalar
    use mcplisr, only : mcpli_potential, mcpli_deriv_second_potential, mcpli_phiend0, &
        mcpli_phi_trajectory, mcpli_deriv_potential, mcpli_InverseVp_primitive, &
        mcpli_phi_trajectory, mcpli_deriv_third_potential, mcpli_Delta_faciq, &
        mcpli_sigma_trajectory
    use curvaton, only : find_MC_bfoldstar, MC7_lnRrad, MC7_Pzeta, MC_Delta, &
                         MC_inftraj_sigma, MC_r, MC_ns, MC7_fNL_phistarfree, &
                         MC7_Pzeta_relativecontribution_sigma, MC_alphaS

    implicit none

    private

    public mcpli7_ns, mcpli7_r, mcpli7_fnl, mcpli7_isnormalizable, mcpli7_M, &
            mcpli7_bfoldstar, mcpli7_Pzeta_sigma, mcpli7_alphaS, mcpli7_sigma_trajectory, mcpli7_ns_faciq, &
           mcpli7_r_faciq, mcpli7_fNL_faciq, mcpli7_alphaS_faciq, mcpli7_phi_trajectory

  contains


!!!!!!!!!!!!!!!!!!!!!!
!! Number of efolds !!
!!!!!!!!!!!!!!!!!!!!!!

!Returns the number of bfolds given potential parameters
  function mcpli7_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli7_bfoldstar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp), parameter :: tonind=tolkp
    real(kp) :: bfoldstarmini,bfoldstarmaxi,bfoldstar
    type(transfert) :: mcpliData

    mcpliData%real1 = Vparam%real1 ! M
    mcpliData%real2 = Vparam%real2 ! alpha
    mcpliData%real3 = msigma
    mcpliData%real4 = Gammaphi
    mcpliData%real5 = Gammasigma
    mcpliData%real6 = sigmaend

    bfoldstarmini = -1000._kp
    bfoldstarmaxi = 0._kp

    bfoldstar = zbrent(find_mcpli7_bfoldstar,bfoldstarmini,bfoldstarmaxi,tonind,mcpliData)

    mcpli7_bfoldstar = bfoldstar

  end function mcpli7_bfoldstar

! Returns the equation one must solve to obtain bfoldstar from model parameters
  function find_mcpli7_bfoldstar(bfoldstar,mcpliData)
    implicit none
    real(kp) :: find_mcpli7_bfoldstar
    real(kp), intent(in) :: bfoldstar
    type(transfert), optional, intent(inout) :: mcpliData
    type(transfert) :: Vparam

    real(kp) :: msigma, Gammaphi, Gammasigma, sigmaend, lnRrad, Pstar, phistar

    Vparam%real1 = mcpliData%real1! M
    Vparam%real2 = mcpliData%real2 ! alpha

    msigma = mcpliData%real3
    Gammaphi = mcpliData%real4
    Gammasigma = mcpliData%real5
    sigmaend = mcpliData%real6

    lnRrad = MC7_lnRrad(mcpli_potential, mcpli_deriv_second_potential, mcpli_phiend0, Vparam, &
             msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    Pstar = MC7_Pzeta(mcpli_potential, mcpli_deriv_potential, mcpli_deriv_second_potential, &
            mcpli_phiend0, mcpli_InverseVp_primitive, Vparam, msigma, Gammaphi, Gammasigma, &
            sigmaend, phistar)

    find_mcpli7_bfoldstar = find_MC_bfoldstar(mcpli_potential, mcpli_deriv_potential, &
         mcpli_deriv_second_potential, mcpli_phiend0, mcpli_InverseVp_primitive, &
         mcpli_phi_trajectory, Vparam, msigma, Gammaphi, Gammasigma, sigmaend, lnRrad, &
         Pstar, bfoldstar)

  end function find_mcpli7_bfoldstar

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Power Spectrum Amplitude  !!
!!      Normalization         !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! Returns the power spectrum amplitude once the number of bfolds_star has been solved
  function mcpli7_Pstar_bfoldssolved(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    real(kp) :: mcpli7_Pstar_bfoldssolved
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: bfoldstar, phistar

    bfoldstar = mcpli7_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    mcpli7_Pstar_bfoldssolved = MC7_Pzeta(mcpli_potential, mcpli_deriv_potential, &
        mcpli_deriv_second_potential,mcpli_phiend0, mcpli_InverseVp_primitive, Vparam, msigma, &
        Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcpli7_Pstar_bfoldssolved

! Returns the equation one must solve to obtain M from other model parameters by normalizing it to the scalar power spectrum amplitude
  function find_mcpli7_M(M,mcpliData)
    implicit none
    real(kp) :: find_mcpli7_M
    real(kp), intent(in) :: M
    type(transfert), optional, intent(inout) :: mcpliData
    real(kp) :: msigma, Gammaphi, Gammasigma, sigmaend, Pstar
    type(transfert) :: Vparam

    Vparam%real1 = M ! M
    Vparam%real2 = mcpliData%real2 ! alpha

    msigma = mcpliData%real3
    Gammaphi = mcpliData%real4
    Gammasigma = mcpliData%real5
    sigmaend = mcpliData%real6

    Pstar = mcpli7_Pstar_bfoldssolved(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)

    find_mcpli7_M = Pstar - powerAmpScalar

  end function find_mcpli7_M

!Returns the mass normalization of the inflaton potential (M) to the power spectrum amplitude
  function mcpli7_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli7_M
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp), parameter :: tonind=tolkp
    real(kp) :: Mmini1, Mmini2, Mmini3, Mmini, Mmaxi, M
    type(transfert) :: mcpliData

    mcpliData%real2 = Vparam%real2 ! alpha
    mcpliData%real3 = msigma
    mcpliData%real4 = Gammaphi
    mcpliData%real5 = Gammasigma
    mcpliData%real6 = sigmaend

    Vparam%real1 = 1._kp ! M (dummy value, to get V/M^4 in Mmini)

    Mmini1 = (msigma**2*sigmaend**2/(3._kp* &
              mcpli_potential(mcpli_phiend0(Vparam),Vparam)))**(0.25_kp) ! This is to have phi dominate the energy budget at the end of inflation
    Mmini2 = (msigma**2/ &
              mcpli_potential(mcpli_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have sigma still light at the end of inflation
    Mmini3 = (2._kp*Gammaphi**2/ mcpli_potential(mcpli_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have Hend>Gammaphi
    Mmini = max(Mmini1,Mmini2,Mmini3)*1.1_kp !1.1 is a safety number
    Mmaxi = 1._kp

    if (find_mcpli7_M(Mmini,mcpliData)*find_mcpli7_M(Mmaxi,mcpliData) > 0._kp) then
         stop 'normalization to Power Spectrum Amplitude is Impossible'
    endif

    M = zbrent(find_mcpli7_M,Mmini,Mmaxi,tonind,mcpliData)

    mcpli7_M = M

  end function mcpli7_M

  !Returns the correlation angle Delta between curvature and isocurvature power spectra
  function mcpli7_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli7_Delta, mphi, bfoldstar
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    type(transfert) :: Vparam_passed

    Vparam_passed%real1 = mcpli7_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) ! Normalizes M
    Vparam_passed%real2 = Vparam%real2 ! passes f
    bfoldstar = mcpli7_bfoldstar(Vparam_passed, msigma, Gammaphi, Gammasigma, sigmaend)
    mcpli7_Delta = MC_Delta(mcpli_potential, mcpli_deriv_potential, &
            mcpli_deriv_second_potential, mcpli_phiend0, mcpli_InverseVp_primitive, &
            mcpli_phi_trajectory, Vparam_passed, msigma, Gammaphi, Gammasigma, &
            sigmaend, bfoldstar)

  end function mcpli7_Delta

!Test Function to determine if the power spectrum is normalizable
  function mcpli7_isnormalizable(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    logical :: mcpli7_isnormalizable
    real(kp), intent(in) :: alpha, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam
    real(kp) :: Mmini1, Mmini2, Mmini3, Mmini, Mmaxi
    type(transfert) :: mcpliData

    Vparam%real1 = 1._kp ! M (dummy value, to get V/M^4 in Mmini)
    Vparam%real2 = alpha

    mcpli7_isnormalizable = .false.
    Mmini1 = (msigma**2*sigmaend**2/(3._kp* &
              mcpli_potential(mcpli_phiend0(Vparam),Vparam)))**(0.25_kp) ! This is to have phi dominate the energy budget at the end of inflation
    Mmini2 = (msigma**2/ &
              mcpli_potential(mcpli_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have sigma still light at the end of inflation
    Mmini3 = (2._kp*Gammaphi**2/ mcpli_potential(mcpli_phiend0(Vparam),Vparam))**(0.25_kp) ! This is to have Hend>Gammaphi
    Mmini = max(Mmini1,Mmini2,Mmini3)*1.1_kp !1.1 is a safety number
    Mmaxi = 1._kp
    mcpliData%real2 = Vparam%real2 ! alpha
    mcpliData%real3 = msigma
    mcpliData%real4 = Gammaphi
    mcpliData%real5 = Gammasigma
    mcpliData%real6 = sigmaend

    if (find_mcpli7_M(Mmini,mcpliData)*find_mcpli7_M(Mmaxi,mcpliData) < 0._kp) then
      mcpli7_isnormalizable = .true.
    endif

  end function mcpli7_isnormalizable

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Observable Predictions   !!
!!    Direct Calculation     !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  function mcpli7_ns(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli7_ns, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: alpha, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = alpha !passes alpha
    Vparam%real1 = mcpli7_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcpli7_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mcpli_potential,mcpli_deriv_second_potential,mcpli_phiend0, &
                mcpli_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mcpli7_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mcpli7_ns = MC_ns(mcpli_potential,mcpli_deriv_potential,mcpli_deriv_second_potential, &
                mcpli_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcpli7_ns

  function mcpli7_alphaS(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli7_alphaS, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: alpha, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = alpha !passes alpha
    Vparam%real1 = mcpli7_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcpli7_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mcpli_potential,mcpli_deriv_second_potential,mcpli_phiend0, &
                mcpli_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mcpli7_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mcpli7_alphaS = MC_alphaS(mcpli_potential,mcpli_deriv_potential,mcpli_deriv_second_potential, &
                mcpli_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcpli7_alphaS

  function mcpli7_r(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli7_r, Delta, bfoldstar, phistar, sigmastar
    real(kp), intent(in) :: alpha, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = alpha !passes alpha
    Vparam%real1 = mcpli7_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcpli7_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    sigmastar = MC_inftraj_sigma(mcpli_potential,mcpli_deriv_second_potential,mcpli_phiend0, &
                mcpli_InverseVp_primitive,Vparam,phistar,msigma,sigmaend)
    Delta = mcpli7_Delta(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    mcpli7_r = MC_r(mcpli_potential,mcpli_deriv_potential,mcpli_deriv_second_potential, &
                Vparam,phistar,sigmastar,msigma,Delta)

  end function mcpli7_r

  function mcpli7_fNL(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli7_fNL, bfoldstar, phistar
    real(kp), intent(in) :: alpha, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = alpha !passes alpha
    Vparam%real1 = mcpli7_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcpli7_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

    mcpli7_fNL=MC7_fNL_phistarfree(mcpli_potential,mcpli_deriv_potential, &
               mcpli_deriv_second_potential, mcpli_phiend0, mcpli_InverseVp_primitive, &
               Vparam, msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcpli7_fNL


! Returns the relative contribution of sigma to the total curvature power spectrum
  function mcpli7_Pzeta_sigma(alpha, msigma, Gammaphi, Gammasigma, sigmaend)
    implicit none
    real(kp) :: mcpli7_Pzeta_sigma, bfoldstar, phistar
    real(kp), intent(in) :: alpha, msigma, Gammaphi, Gammasigma, sigmaend
    type(transfert) :: Vparam

    Vparam%real2 = alpha !passes alpha
    Vparam%real1 = mcpli7_M(Vparam, msigma, Gammaphi, Gammasigma, sigmaend) !passes M
    bfoldstar = mcpli7_bfoldstar(Vparam, msigma, Gammaphi, Gammasigma, sigmaend)
    phistar = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

    mcpli7_Pzeta_sigma=MC7_Pzeta_relativecontribution_sigma(mcpli_potential,mcpli_deriv_potential, &
                      mcpli_deriv_second_potential,mcpli_phiend0,mcpli_InverseVp_primitive,Vparam, &
                      msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcpli7_Pzeta_sigma

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Observable Predictions   !!
!! Step by Step Calculation  !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!returns phistar once bfoldstar is computed
  function mcpli7_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)
    implicit none
    real(kp), intent(in) :: bfoldstar, msigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: mcpli7_phi_trajectory

    mcpli7_phi_trajectory = mcpli_phi_trajectory(bfoldstar, Vparam, msigma, sigmaend)

  end function mcpli7_phi_trajectory

!returns sigmastar once phistar is computed
  function mcpli7_sigma_trajectory(phistar, Vparam, msigma, sigmaend)
    implicit none
    real(kp), intent(in) :: phistar, msigma, sigmaend
    type(transfert), optional, intent(inout) :: Vparam
    real(kp) :: mcpli7_sigma_trajectory

    mcpli7_sigma_trajectory = mcpli_sigma_trajectory(phistar, Vparam, msigma, sigmaend)

  end function mcpli7_sigma_trajectory

!returns ns from already computed intermediary quantites (faciq)
  function mcpli7_ns_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcpli7_ns_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcpli_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcpli7_ns_faciq = MC_ns(mcpli_potential,mcpli_deriv_potential,mcpli_deriv_second_potential, &
                mcpli_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcpli7_ns_faciq

!returns r from already computed intermediary quantites (faciq)
  function mcpli7_r_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcpli7_r_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcpli_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcpli7_r_faciq = MC_r(mcpli_potential,mcpli_deriv_potential,mcpli_deriv_second_potential, &
                Vparam,phistar,sigmastar,msigma,Delta)

  end function mcpli7_r_faciq

!returns fNL from already computed intermediary quantites (faciq)
  function mcpli7_fNL_faciq(msigma, Gammaphi, Gammasigma, sigmaend, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcpli7_fNL_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, sigmaend, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcpli_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcpli7_fNL_faciq = MC7_fNL_phistarfree(mcpli_potential,mcpli_deriv_potential, &
               mcpli_deriv_second_potential, mcpli_phiend0, mcpli_InverseVp_primitive, &
               Vparam, msigma, Gammaphi, Gammasigma, sigmaend, phistar)

  end function mcpli7_fNL_faciq

!returns alphaS from already computed intermediary quantites (faciq)
  function mcpli7_alphaS_faciq(msigma, Gammaphi, Gammasigma, Vparam, &
                        phistar, sigmastar)
    implicit none
    real(kp) :: mcpli7_alphaS_faciq, Delta
    real(kp), intent(in) :: msigma, Gammaphi, Gammasigma, phistar, sigmastar
    type(transfert), intent(in) :: Vparam

    Delta = mcpli_Delta_faciq(Vparam, msigma, phistar, sigmastar)
    mcpli7_alphaS_faciq = MC_alphaS(mcpli_potential,mcpli_deriv_potential,mcpli_deriv_second_potential, &
                mcpli_deriv_third_potential,Vparam,phistar,sigmastar,msigma,Delta)

  end function mcpli7_alphaS_faciq





  end module mcpli7reheat
